import arrow
import copy

from abc import ABCMeta, abstractmethod

from ddd.utils.utils import get_for_compare


class Event(object, metaclass=ABCMeta):
    """
    An event base class.
    """
    def __init__(self, name, version="1", date=None, sender=None, corr_ids=None):
        date = date if date is not None else arrow.utcnow()
        corr_ids = corr_ids if corr_ids is not None else []

        self.name = name
        self.version = version
        self.date = date
        self.sender = sender
        self.corr_ids = corr_ids

    def equals(self, other, ignore_fields):
        ignore_fields = ignore_fields if ignore_fields is not None else []

        if isinstance(other, self.__class__):

            a = get_for_compare(self, ignore_fields)
            b = get_for_compare(other, ignore_fields)

            return a == b
        else:
            return False

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            a = copy.deepcopy(self.__dict__)
            b = copy.deepcopy(other.__dict__)
            a.pop('date')
            b.pop('date')
            return a == b
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @abstractmethod
    def get_serialized_payload(self):
        """
        Implemented by subclasses to provide the payload.
        """
        pass

    def serialize(self):
        obj = {
            'name': self.name,
            'version': self.version,
            'date': self.date.isoformat(),
            'sender': self.sender,
            'payload': self.get_serialized_payload(),
            'corr_ids': self.corr_ids,
        }
        return obj



from abc import abstractmethod


class JobAdapter(object):
    """
    The secondary 'scheduler' adapter schedules jobs.

    This is the primary 'jobs' adapter that is responsible for handling jobs
    when they are executed.
    """

    def __init__(self, log_service, token):
        super().__init__()
        self.set_log_service(log_service)
        self.set_token(token)

    # Setters

    @abstractmethod
    def set_log_service(self, log_service):
        pass

    @abstractmethod
    def set_token(self, token):
        pass

    @abstractmethod
    def set_service(self, service):
        pass

    @classmethod
    async def run_job(cls, **kwargs):
        """
        Does nothing since application logic (such as jobs)
        should be in your domain application service.

        Please override this class.
        """
        raise Exception(
            "The base class run_job method was called. "
            "This should never happen. You must have scheduled a job "
            "without having implemented a job adapter to handle the job. "
            "Please override the class if you need jobs."
        )



"""
The http adapter exceptions.
"""
class HttpException(Exception):
    """
    Base class for http adapter exceptions.
    This class must not be used directly. Throw one of it's subclasses instead.
    """
    def __init__(self, exception):
        super(HttpException, self).__init__()
        self.exception = exception


class NotReachableException(HttpException):
    def __init__(self, exception):
        super(NotReachableException, self).__init__(exception=exception)



from ddd.domain.aggregate import Aggregate


class Dummy(Aggregate):
    """
    A 'dummy' for tests.
    """
    VERSION = "2"

    def __init__(
        self,
        version,
        dummy_id,
        property_added,
    ):
        super().__init__(version=version)
        self.dummy_id = dummy_id
        self.property_added = property_added



from abc import ABCMeta, abstractmethod


class HttpAdapter(object, metaclass=ABCMeta):
    def __init__(self, config, loop, log_service, webhook_callback_token):
        super(HttpAdapter, self).__init__()

        self.config = config
        self.loop = loop
        self.log_service = log_service
        self.webhook_callback_token = webhook_callback_token

    # Abstract methods (must be implemented by superclasses)

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass

    # Service

    def set_service(self, service):
        self.service = service



import asyncio
import json

from azure.eventhub.aio import EventHubConsumerClient
from azure.eventhub.extensions.checkpointstoreblobaio import \
    BlobCheckpointStore

from ddd.adapters.event.event_adapter import EventAdapter


class AzureEventAdapter(EventAdapter):

    def __init__(
        self,
        namespace_conn_string,
        checkpoint_store_conn_string,
        blob_container_name,
        topic,
        group,
        log_service,
        service=None,
        listeners=None,
    ):
        super().__init__(
            log_service=log_service,
            service=service,
            listeners=listeners
        )

        self.namespace_conn_string = namespace_conn_string
        self.checkpoint_store_conn_string = checkpoint_store_conn_string
        self.blob_container_name = blob_container_name
        self.topic = topic
        self.group = group

        self.checkpoint_store = None
        self.consumer = None
        self.consumer_task = None

    # Handling

    async def handle(self, message):
        """
        Handle a message.
        """
        message = json.loads(message.body_as_str(encoding='UTF-8'))
        await super().handle(message)

    # Control

    async def start(self):
        await self._create_checkpoint_store()
        await self._create_consumer()
        await self._start_consuming()

    async def stop(self):
        await self._stop_consuming()

    # Azure

    async def _create_checkpoint_store(self):
        self.checkpoint_store = \
            BlobCheckpointStore.from_connection_string(
                self.checkpoint_store_conn_string,
                self.blob_container_name,
            )

    async def _create_consumer(self):
        self.consumer = \
            EventHubConsumerClient.from_connection_string(
                self.namespace_conn_string,
                consumer_group=self.group,
                eventhub_name=self.topic,
                checkpoint_store=self.checkpoint_store,
            )

    async def _start_consuming(self):
        self.consumer_task = \
            asyncio.ensure_future(
                self.consumer.receive(
                    on_event=self._on_event,
                )
            )
        await asyncio.sleep(0.01)

    async def _on_event(self, partition_context, event):
        try:
            await self.handle(event)
        except Exception as e:
            self.log_service.error(
                f"Azure event adapter got exception when "
                f"delegating call to service: '{str(e)}'.",
                exc_info=True,
            )

        # Update the checkpoint so that the program doesn't read the events
        # that it has already read when you run it next time.
        await partition_context.update_checkpoint(event)

    async def _stop_consuming(self):
        self.consumer_task.cancel()
        await self.consumer.close()



from ddd.repositories.postgres.postgres_repository import PostgresRepository

from ddd.domain.dummy.dummy import Dummy
from ddd.domain.dummy.dummy_repository import DummyRepository
from ddd.domain.dummy.dummy_translator import DummyTranslator


class PostgresDummyRepository(DummyRepository, PostgresRepository):
    def __init__(
        self,
        config,
        db_service,
        log_service,
        loop=None,
    ):
        PostgresRepository.__init__(
            self,
            loop=loop,
            db_service=db_service,
            log_service=log_service,
            aggregate_cls=Dummy,
            translator_cls=DummyTranslator,
            table_name="ddd_dummies",
            dsn=config.database.postgres.dsn,
        )

    # Operations

    async def get(self, dummy_id):
        return await self._get(aggregate_id=dummy_id)

    async def save(self, dummy):
        await self._save(
            aggregate_id=dummy.dummy_id,
            aggregate=dummy,
        )

    async def delete(self, dummy):
        await self._delete(aggregate_id=dummy.dummy_id)





class Command(object):
    def __init__(self):
        super(Command, self).__init__()

    # Equality

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)



from ddd.domain.value_object import ValueObject


class EntityId(ValueObject):
    def __init__(self, identity, version="1"):
        super(EntityId, self).__init__(version=version)
        self.identity = str(identity)

    def __str__(self):
        return self.identity

    def __eq__(self, other):
        if not isinstance(other, EntityId):
            return self.identity == str(other)

        return self.identity == other.identity



class DomainRegistry(object):
    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if DomainRegistry.__instance is None:
            DomainRegistry()
        return DomainRegistry.__instance

    def __init__(self):
        """ Virtually private constructor. """
        if DomainRegistry.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            DomainRegistry.__instance = self
            DomainRegistry.__instance.domain_publisher = None
            DomainRegistry.__instance.interchange_publisher = None
            DomainRegistry.__instance.task_db_conns = {}
            DomainRegistry.__instance.event_repository = None
            DomainRegistry.__instance.service = None
            DomainRegistry.__instance.log_service = None



"""
The job adapter exceptions.
"""
class JobException(Exception):
    """
    Base class.
    This class must not be used directly.
    """
    def __init__(self, message):
        super().__init__(message)


class JobNotFound(JobException):
    def __init__(self, job_id):
        super().__init__(
            f"Couldn't find job with ID: '{job_id}'."
        )
        self.job_id = job_id



from abc import ABCMeta, abstractmethod

from ddd.infrastructure.infrastructure_service import InfrastructureService


class DbService(InfrastructureService, metaclass=ABCMeta):
    """
    A db service base class.
    """
    def __init__(
        self,
        log_service,
    ):
        super().__init__(log_service=log_service)
        self.conn_pool = None

    @abstractmethod
    async def _create_conn_pool(self):
        pass

    async def start(self):
        self.log_service.info("..starting db service")
        await self._create_conn_pool()

    async def stop(self):
        self.log_service.info("..stopping db service")
        await self.conn_pool.close()



import os

from unittest.async_case import IsolatedAsyncioTestCase

from abc import abstractmethod

from dotenv import find_dotenv, load_dotenv


class BaseTestCase(IsolatedAsyncioTestCase):

    def __init__(self, env_file_path=None, methodName='runTest'):
        super().__init__(
            methodName=methodName
        )

        if env_file_path in [None, ""]:
            env_file_path = \
                find_dotenv(
                    filename=os.getenv('ENV_FILE'),
                    raise_error_if_not_found=True,
                    usecwd=True,
                )

        self.env_file_path = env_file_path

    async def asyncSetUp(self):
        await super().asyncSetUp()

        # Vars
        self.config = None
        self.deps = None
        self.loop = None

        # Load env vars
        load_dotenv(
            dotenv_path=self.env_file_path
        )

        # Read config
        self.read_config()

    async def asyncTearDown(self):
        await super().asyncTearDown()

    @abstractmethod
    def read_config(self):
        """
        Read the config into 'self.config'.
        """
        pass

    # Assert

    def assertEqualIgnoringId(self, id_field, a, b):
        """
        Convenience method for 'assertEqualIgnoringIds'.
        """
        self.assertEqualIgnoringIds(
            id_fields=[id_field],
            a=a,
            b=b,
        )

    def assertEqualIgnoringIds(self, id_fields, a, b):
        """
        This is an assertion method used to assert two entities
        are equal, disregarding some entitiy IDs.
        """
        self.assertEqualIgnoringFields(
            ignore_fields=id_fields,
            a=a,
            b=b,
        )

    def assertEqualIgnoringFields(self, a, b, ignore_fields=None):
        """
        This is an assertion method used to assert two entities
        are equal, disregarding some entitiy IDs.
        """
        ignore_fields = ignore_fields if ignore_fields is not None else []

        if type(a) == list:
            self.assertListsEqualIgnoringFields(a, b, ignore_fields)
        else:
            if not a.equals(b, ignore_fields=ignore_fields):
                self.assertEqual(
                    a.serialize(ignore_fields=ignore_fields),
                    b.serialize(ignore_fields=ignore_fields),
                    "The two building blocks didn't match.",
                )

    def assertListsEqualIgnoringFields(self, a, b, ignore_fields=None):
        """
        The lists equals.
        """
        ignore_fields = ignore_fields if ignore_fields is not None else []

        from ddd.utils.utils import get_for_compare

        a = get_for_compare(a, ignore_fields)
        b = get_for_compare(b, ignore_fields)

        if not a == b:
            self.assertEqual(
                a,
                b,
                "The two lists didn't match.",
            )

    # Helpers

    @classmethod
    def Any(cls, type_=None):
        """
        Returns an object that can be used to compare with other objects.

        Example:
            Any(str) == some_var
            Any(float) == some_var
        """

        class Any(cls):
            def __eq__(self, other):
                if type_ is None:
                    return other is not None
                return type(other) == type_

        return Any()



from unittest import skip

from ddd.tests.dummy_action_test_case import DummyActionTestCase

from ddd.adapters.event.memory.memory_event_adapter import MemoryEventAdapter

from shipping.adapters.listeners.domain.shipment_created_listener \
    import ShipmentCreatedListener
from shipping.application.shipping_application_service import \
    ShippingApplicationService
from shipping.domain.commands import SendTrackingEmailCommand


class TestEventAdapter(DummyActionTestCase):

    async def asyncSetUp(self):
        await super().asyncSetUp()

        self.adapter = MemoryEventAdapter()

    async def test_assigns_service_to_listeners_when_setting_service(
        self,
    ):
        # Setup
        listener = \
            ShipmentCreatedListener(
                action="send_tracking_email",
                service=None,
                command_creator=lambda event:
                SendTrackingEmailCommand(
                    shipment_id=event.shipment_id,
                    token=None,
                )
            )

        self.adapter.set_listeners([
            listener
        ])

        service = ShippingApplicationService(
            customer_repository=None,
            db_service=None,
            domain_adapter=None,
            domain_publisher=None,
            event_repository=None,
            interchange_adapter=None,
            interchange_publisher=None,
            job_adapter=None,
            job_service=None,
            log_service=None,
            scheduler_adapter=None,
            shipment_repository=None,
            max_concurrent_actions=10,
            loop=None,
        )

        # Exercise
        self.adapter.set_service(service)

        # Assert
        for listener in self.adapter.listeners:
            self.assertEqual(service, listener.service)

    @skip
    async def test_assigns_service_to_listeners_when_setting_listeners(
        self,
    ):
        raise NotImplementedError()

    @skip
    async def test_delegates_event_to_listeners(
        self,
    ):
        raise NotImplementedError()



from ddd.infrastructure.infrastructure_service import InfrastructureService

from abc import ABCMeta, abstractmethod


class JobService(InfrastructureService, metaclass=ABCMeta):
    """
    A job service base class.
    """
    def __init__(
        self,
        scheduler_adapter,
        log_service,
    ):
        super().__init__(log_service=log_service)
        self.scheduler_adapter = scheduler_adapter

    async def start(self):
        await self.scheduler_adapter.start()

    async def stop(self):
        await self.scheduler_adapter.stop()

    # Schedule jobs

    async def add_cron_job(
        self,
        id,
        name,
        args,
        kwargs,
        weekdays,
        hour,
        minute,
        second,
        start_date,
        end_date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        return self.scheduler_adapter.add_cron_job(
            id=id,
            name=name,
            args=args,
            kwargs=kwargs,
            weekdays=weekdays,
            hour=hour,
            minute=minute,
            second=second,
            start_date=start_date.datetime,
            end_date=end_date.datetime,
            timezone=timezone,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            replace_existing=replace_existing,
        )

    async def add_date_job(
        self,
        id,
        name,
        args,
        kwargs,
        date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        return self.scheduler_adapter.add_date_job(
            id=id,
            name=name,
            args=args,
            kwargs=kwargs,
            date=date.datetime,
            timezone=timezone,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            replace_existing=replace_existing,
        )

    async def add_interval_job(
        self,
        id,
        name,
        args,
        kwargs,
        interval,
        days,
        hours,
        minutes,
        seconds,
        start_date,
        end_date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        return await self.scheduler_adapter.add_interval_job(
            id=id,
            name=name,
            args=args,
            kwargs=kwargs,
            interval=interval,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            start_date=start_date.datetime,
            end_date=end_date.datetime,
            timezone=timezone,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            replace_existing=replace_existing,
        )

    async def get_all_jobs(self):
        return await self.scheduler_adapter.get_all_jobs()

    async def get_job_count(self):
        return await self.scheduler_adapter.get_job_count()

    async def remove_job(self, job_id, raises=False):
        await self.scheduler_adapter.remove_job(
            job_id=job_id,
            raises=raises,
        )

    async def remove_all_jobs(self):
        await self.scheduler_adapter.remove_all_jobs()



import asyncio
import json

from aiokafka import AIOKafkaConsumer

from ddd.adapters.event.event_adapter import EventAdapter


class KafkaEventAdapter(EventAdapter):

    def __init__(
        self,
        bootstrap_servers,
        topic,
        group,
        log_service,
        service=None,
        listeners=None,
    ):
        super().__init__(
            log_service=log_service,
            service=service,
            listeners=listeners
        )

        self.bootstrap_servers = bootstrap_servers
        self.topic = topic
        self.group = group

        self.consumer_connected = False
        self.consumer_stopped = False
        self.consuming = False

    # Handling

    async def handle(self, message):
        """
        Handle a message.
        """
        message = json.loads(message.value)
        await super().handle(message)

    # Control

    async def start(self):
        await self._create_consumer()
        await self._connect()
        await self._start_consuming()

    async def stop(self):
        await self._stop_consuming()
        await self._disconnect()

    async def _create_consumer(self):
        self.consumer = \
            AIOKafkaConsumer(
                self.topic,
                loop=asyncio.get_event_loop(),
                bootstrap_servers=self.bootstrap_servers,
                group_id=self.group,
                enable_auto_commit=False
            )

    async def _connect(self):
        created = False
        error = None
        backoff = 6
        retries = 0
        max_retries = 3

        while (not created) and (retries < max_retries):
            try:
                await self.consumer.start()
                self.consumer_connected = True
                created = True

            except Exception as e:
                self.log_service.error(
                    (
                        "{} failed to connect to: {} "
                        "(retrying in {} secs..), exception: {}"
                    ).format(
                        "Kafka Event Adapter",
                        self.bootstrap_servers,
                        backoff,
                        str(e)
                    )
                )
                error = str(e)
                retries += 1
                await asyncio.sleep(backoff)

        if not created:
            raise Exception(
                (
                    "{} failed to connect to: {} "
                    "after {} retries, error: {}."
                ).format(
                    "Kafka Event Adapter",
                    self.bootstrap_servers,
                    retries,
                    error
                )
            )

        await self._wait_for_connect()

        self.log_service.debug(
            "Kafka event adapter listens on {} @ {}, (group: {})".
                format(
                self.topic,
                self.bootstrap_servers,
                self.group
            )
        )

    async def _wait_for_connect(self, timeout=60):
        waited = 0
        while not self.consumer_connected:
            await asyncio.sleep(1)
            waited = waited + 1
            if waited > timeout:
                raise Exception(
                    "Couldn't connect to kafka, timed out after {} secs".
                    format(
                        timeout
                    )
                )

    async def _start_consuming(self):
        # Schedule in a new task on the event loop to prevent blocking
        asyncio.get_event_loop().create_task(
            self._consume()
        )

    async def _consume(self):
        self.consuming = True

        while not self.consumer_stopped:
            try:
                message = await asyncio.wait_for(
                    self.consumer.__anext__(),
                    timeout=3
                )
                try:
                    await self.handle(message)
                except Exception as e:
                    self.log_service.error(
                        f"Kafka Event Adapter got exception when "
                        f"delegating call to service: '{str(e)}'.",
                        exc_info=True,
                    )
                await self.consumer.commit()
            except asyncio.TimeoutError:
                pass

        self.consuming = False

    async def _stop_consuming(self):
        self.consumer_stopped = True
        await self.wait_for_consumption_to_stop()

    async def wait_for_consumption_to_stop(self, timeout=60):
        waited = 0
        while self.consuming:
            await asyncio.sleep(1)
            waited = waited + 1
            if waited > timeout:
                raise Exception(
                    (
                        "Failed waiting for consumption to stop, "
                        "timed out after {} secs"
                    ).format(
                        timeout
                    )
                )

    async def _disconnect(self):
        # Will leave consumer group; perform autocommit if enabled.
        await self.consumer.stop()
        self.consumer_connected = False

    async def _wait_for_disconnect(self, timeout=60):
        waited = 0
        while self.consumer_connected:
            await asyncio.sleep(1)
            waited = waited + 1
            if waited > timeout:
                raise Exception(
                    (
                        "Couldn't disconnect from kafka, timed out "
                        "after {} secs"
                    ).format(
                        timeout
                    )
                )



import asyncio
import json

from aiokafka import AIOKafkaProducer

from ddd.domain.event_publisher import EventPublisher


class KafkaEventPublisher(EventPublisher):
    """
    A kafka event publisher.
    """
    def __init__(
        self,
        bootstrap_servers,
        topic,
        group,
        log_service,
        keep_flushed_copies=False,
    ):
        super().__init__(
            log_service=log_service,
            keep_flushed_copies=keep_flushed_copies
        )
        self.bootstrap_servers = bootstrap_servers
        self.topic = topic
        self.group = group

        self.is_started = False

    # Flush

    async def flush(self, event):
        """
        Actually publishes the event,
        (called after request is done in the 'action' decorator).
        """
        task_id = self._get_loop_task_id()

        prefix = "[ -------- ]"

        if not self.producer:
            raise Exception(
                (
                    "Kafka event publisher couldn't publish because "
                    "producer has not been created."
                )
            )

        await self.producer.send_and_wait(
            self.topic,
            json.dumps(event.serialize()).encode()
        )

        if self.keep_flushed_copies:
            if task_id not in self.flushed:
                self.flushed[task_id] = []

            self.flushed[task_id].append(event)

    # Control

    async def start(self):
        await self.create_producer()
        await self.connect(
            retry_backoff=6,
            max_retries=3,
        )

    async def stop(self):
        if not self.producer:
            self.log_service.warning(
                (
                    "Kafka event publisher was instructed to stop "
                    "but it doesn't seem to be started (no producer)."
                )
            )
        else:
            await self.disconnect()

    # Connect

    async def connect(self, retry_backoff=6, max_retries=12):
        created = False
        error = None
        retry_backoff = retry_backoff
        retries = 0
        max_retries = max_retries

        while (not created) and retries < max_retries:
            try:
                await self.producer.start()
                self.producer_connected = True
                created = True

            except Exception as e:
                self.log_service.error(
                    (
                        "{} failed to connect to: {} "
                        "(retrying in {} secs..), exception: {}"
                    ).
                    format(
                        "Kafka Event Publisher",
                        self.bootstrap_servers,
                        retry_backoff,
                        str(e)
                    )
                )
                error = str(e)
                retries += 1
                await asyncio.sleep(retry_backoff)

        if not created:
            raise Exception(
                (
                    "{} failed to connect to: {} "
                    "after {} retries, error: {}."
                ).
                format(
                    "Kafka Event Publisher",
                    self.bootstrap_servers,
                    retries,
                    error
                )
            )

        await self.wait_for_connect()

        self.log_service.debug(
            "Kafka event publisher connected to {} @ {}".
            format(
                self.topic,
                self.bootstrap_servers
            )
        )

    async def disconnect(self):
        await self.producer.stop()
        self.producer_connected = False

    async def wait_for_connect(self, timeout=60):
        waited = 0
        while not self.producer_connected:
            await asyncio.sleep(1)
            waited = waited + 1
            if waited > timeout:
                raise Exception(
                    "Couldn't connect to kafka, timed out after {} secs".
                    format(
                        timeout
                    )
                )

    # Publishing

    async def create_producer(self):
        self.producer = AIOKafkaProducer(
            loop=asyncio.get_event_loop(),
            bootstrap_servers=self.bootstrap_servers
        )



from abc import ABCMeta, abstractmethod


class InfrastructureService(object, metaclass=ABCMeta):
    """
    A infrastructure service base class.
    """
    def __init__(self, log_service):
        super().__init__()

        self.log_service = log_service

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass



from ddd.adapters.event.event_adapter import EventAdapter


class MemoryEventAdapter(EventAdapter):

    def __init__(
        self,
        log_service=None,
        service=None,
        listeners=None,
    ):
        super().__init__(
            log_service=log_service,
            service=service,
            listeners=listeners
        )

    # Control

    async def start(self):
        pass

    async def stop(self):
        pass



from ddd.tests.dummy_action_test_case import DummyActionTestCase
from ddd.utils.tasks.migrate_models import Task as MigrateModelsTask
from ddd.domain.dummy.dummy import Dummy


class TestMigrateModels(DummyActionTestCase):

    async def asyncSetUp(self):
        await super().asyncSetUp()

    async def test_records_migrated_to_latest_version_when_running(
        self,
    ):
        # Setup
        Dummy.VERSION = "2"

        aggregate_id = "some-dummy-id"

        data = {
            'version': "1",
            'dummy_id': aggregate_id,
        }

        repository = self.deps.get_dummy_repository()

        await repository._save_record(
            aggregate_id=aggregate_id,
            data=data,
        )

        command = \
            MigrateModelsTask(
                config=self.config,
                deps_mgr=self.deps,
                args_str="",
            )

        # Exercise
        await command._run()

        record = \
            await repository._get_record(
                aggregate_id=aggregate_id,
            )

        # Assert
        self.assertEqual(
            {
                'id': aggregate_id,
                'data': {
                    'version': "2",
                    'dummy_id': aggregate_id,
                    'property_added': 'in_version_2',
                },
            },
            dict(record),
        )



from apscheduler.executors.asyncio import AsyncIOExecutor
from apscheduler.jobstores.base import JobLookupError
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger

from ddd.adapters.job.exceptions import JobNotFound
from ddd.adapters.job.scheduler_adapter import SchedulerAdapter


class ApSchedulerAdapter(SchedulerAdapter):
    def __init__(self, dsn, exec_func, log_service):
        super().__init__(
            exec_func=exec_func,
            log_service=log_service
        )
        self.dsn = dsn

        self.scheduler = None

        self.create_scheduler()

    def create_scheduler(self):
        self.scheduler = AsyncIOScheduler()

        self.scheduler.configure(
            jobstores={
                'default': SQLAlchemyJobStore(url=self.dsn),
            },
            executors={
                'default': AsyncIOExecutor(),
            },
            job_defaults={
                'coalesce': False,
                'max_instances': 1,
                'misfire_grace_time': (60 * 60)
            },
            timezone="UTC",
        )

    async def start(self):
        self.scheduler.start()

    async def stop(self):
        self.scheduler.shutdown(wait=True)

    async def add_cron_job(
        self,
        id,
        name,
        args,
        kwargs,
        crontab_expr,
        weekdays,
        hour,
        minute,
        second,
        start_date,
        end_date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        weekdays = weekdays if weekdays is not None else []

        trigger = None

        if crontab_expr is not None:
            trigger = \
                CronTrigger.from_crontab(
                    crontab_expr,
                    timezone=timezone,
                )
        else:
            trigger = \
                CronTrigger(
                    day_of_week=
                    ",".join(
                        [w[0:3].upper() for w in weekdays]
                    )
                    if len(weekdays) > 0
                    else None,
                    hour=hour,
                    minute=minute,
                    second=second,
                    start_date=start_date,
                    end_date=end_date,
                    timezone=timezone,
                )

        await self._add_job(
            id=id,
            name=name,
            func=self.exec_func,
            kwargs=kwargs,
            trigger=trigger,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            replace_existing=replace_existing,
        )

    async def add_date_job(
        self,
        id,
        name,
        args,
        kwargs,
        date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        raise NotImplementedError()

    async def add_interval_job(
        self,
        id,
        name,
        args,
        kwargs,
        interval,
        days,
        hours,
        minutes,
        seconds,
        start_date,
        end_date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        raise NotImplementedError()

    async def get_all_jobs(self):
        return self.scheduler.get_jobs()

    async def get_job(self, job_id):
        return self.scheduler.get_job(job_id)

    async def get_job_count(self):
        return len(self.scheduler.get_jobs())

    async def remove_job(self, job_id, raises=False):
        try:
            self.scheduler.remove_job(job_id)
        except JobLookupError as e:
            if "No job by the id of" in str(e):
                self.log_service.error(
                    f"Tried to remove apscheduler job with ID '{job_id}' but "
                    f"it wasn't found. Ignoring but you should look this "
                    f"up since it should never happen."
                )
                if raises:
                    raise JobNotFound(job_id)
            else:
                raise

    async def remove_all_jobs(self):
        """
        Removes all scheduled jobs.
        """
        for job in self.scheduler.get_jobs():
            self.scheduler.remove_job(job.id)

    # Helpers

    async def _add_job(self, **kwargs):
        """
        Convenience method for adding a job.
        """
        self.scheduler.add_job(
            **kwargs
        )



class DomainException(Exception):
    """
    A domain exception.
    """
    def __init__(self, errors):
        super(DomainException, self).__init__()
        self.errors = errors

    def __str__(self):
        return ",".join([e['message'] for e in self.errors])



import asyncio
import uvloop

from abc import ABCMeta, abstractmethod

from ddd.utils.dep_mgr import DependencyManager
from ddd.tests.base_test_case import BaseTestCase


class ActionTestCase(BaseTestCase, metaclass=ABCMeta):

    def __init__(self, env_file_path=None, methodName='runTest'):
        super().__init__(
            env_file_path=env_file_path,
            methodName=methodName,
        )

        # Configure
        self.maxDiff = None

    async def asyncSetUp(self):
        await super().asyncSetUp()

        # Vars
        self.service = None

        # Loop
        self.loop = uvloop.new_event_loop()
        asyncio.set_event_loop(self.loop)

        # Deps manager
        self._create_deps_manager()

        # Register service
        self.deps.set_service(self.get_service())

        # Start service
        await self.get_service().start()

    async def asyncTearDown(self):
        await super().asyncTearDown()

        # TODO: Delete all events in events repository, when not memory rep?

        await self.get_service().stop()

    # Dependencies

    def _create_deps_manager(self):
        self.deps = DependencyManager(
            config=self.config,
            loop=self.loop,
        )

    @abstractmethod
    def get_service(self):
        pass

    # Events

    async def simulate_domain_event(self, event):
        await self.deps.get_domain_publisher().flush(event=event)

        # In case we run test with kafka or azure, we need delay
        # to allow for delivery.
        if self.config.pubsub.domain.provider == "kafka":
            await asyncio.sleep(0.01)
        elif self.config.pubsub.domain.provider == "azure":
            await asyncio.sleep(1.5)

    async def simulate_integration_event(self, event):
        await self.deps.get_interchange_publisher().flush(event=event)

        # In case we run test with kafka or azure, we need delay
        # to allow for delivery.
        if self.config.pubsub.interchange.provider == "kafka":
            await asyncio.sleep(0.01)
        elif self.config.pubsub.interchange.provider == "azure":
            await asyncio.sleep(1.5)


    # Assert

    def assertDomainEventPublished(self, event, ignore_fields=None):
        """
        Asserts that the 'event' domain event was published.
        """
        ignore_fields = ignore_fields if ignore_fields is not None else []

        if 'date' not in ignore_fields:
            ignore_fields.append('date')

        self.assertTrue(
            self.deps.get_domain_publisher().has_flushed(
                event=event,
                ignore_fields=ignore_fields
            ),
            f"Expected domain event to "
            f"have been published: '{event.name}'"
        )

    def assertDomainEventNotPublished(self, event, ignore_fields=None):
        """
        Asserts that the 'event' domain event was NOT published.
        """
        ignore_fields = ignore_fields if ignore_fields is not None else []

        if 'date' not in ignore_fields:
            ignore_fields.append('date')

        self.assertFalse(
            self.deps.get_domain_publisher().has_flushed(
                event=event,
                ignore_fields=ignore_fields
            ),
            f"Expected domain event to NOT "
            f"have been published: '{event.name}'"
        )

    def assertIntegrationEventPublished(self, event, ignore_fields=None):
        """
        Asserts that the 'event' integration event was published.
        """
        ignore_fields = ignore_fields if ignore_fields is not None else []

        if 'date' not in ignore_fields:
            ignore_fields.append('date')

        self.assertTrue(
            self.deps.get_interchange_publisher().has_flushed(
            event=event,
                ignore_fields=ignore_fields
            ),
            f"Expected integration event to "
            f"have been published: '{event.name}'"
        )



from abc import ABCMeta, abstractmethod


class DummyRepository(object, metaclass=ABCMeta):
    def __init__(
        self,
    ):
        DummyRepository.__init__(self)

    # Operations



    # Migration

    def _migrate_v1_to_v2(self, record):
        """
        Add: 'property_added'
        """
        record['data']['property_added'] = "in_version_2"

        return record



from ddd.adapters.job.scheduler_adapter import SchedulerAdapter


class MemorySchedulerAdapter(SchedulerAdapter):
    def __init__(self, exec_func, log_service):
        super().__init__(
            exec_func=exec_func,
            log_service=log_service
        )

        self.jobs = {
            'cron': {},
            'date': {},
            'interval': {},
        }

    async def start(self):
        pass

    async def stop(self):
        pass

    async def add_cron_job(
        self,
        id,
        name,
        args,
        kwargs,
        crontab_expr,
        weekdays,
        hour,
        minute,
        second,
        start_date,
        end_date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        self.jobs['cron'][id] = {
            'id': id,
            'name': name,
            'args': args,
            'kwargs': kwargs,
            'crontab_expr': crontab_expr,
            'weekdays': weekdays,
            'hour': hour,
            'minute': minute,
            'second': second,
            'start_date': start_date,
            'end_date': end_date,
            'timezone': timezone,
            'coalesce': coalesce,
            'misfire_grace_time': misfire_grace_time,
        }

    async def add_date_job(
        self,
        id,
        name,
        args,
        kwargs,
        date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        self.jobs['date'][id] = {
            'id': id,
            'name': name,
            'args': args,
            'kwargs': kwargs,
            'date': date,
            'timezone': timezone,
            'coalesce': coalesce,
            'misfire_grace_time': misfire_grace_time,
        }

    async def add_interval_job(
        self,
        id,
        name,
        args,
        kwargs,
        interval,
        days,
        hours,
        minutes,
        seconds,
        start_date,
        end_date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        self.jobs['interval'][id] = {
            'id': id,
            'name': name,
            'args': args,
            'kwargs': kwargs,
            'interval': interval,
            'days': days,
            'hours': hours,
            'minutes': minutes,
            'seconds': seconds,
            'start_date': start_date,
            'end_date': end_date,
            'timezone': timezone,
            'coalesce': coalesce,
            'misfire_grace_time': misfire_grace_time,
        }

    async def get_all_jobs(self):
        jobs = []

        for type_, js in self.jobs.items():
            for job_id, job in js.items():
                jobs.append(job)

        return jobs

    async def get_job(self, job_id):
        for i, j in self.jobs['cron'].items():
            if i == job_id:
                return j

        for i, j in self.jobs['date'].items():
            if i == job_id:
                return j

        for i, j in self.jobs['interval'].items():
            if i == job_id:
                return j

    async def get_job_count(self):
        return \
            len(self.jobs['cron']) + \
            len(self.jobs['date']) + \
            len(self.jobs['interval'])

    async def remove_job(self, job_id, raises=False):
        self.jobs['cron'] = {
            i: j
            for i, j in self.jobs['cron'].items()
            if i != job_id
        }

        self.jobs['date'] = {
            i: j
            for i, j in self.jobs['date'].items()
            if i != job_id
        }

        self.jobs['interval'] = {
            i: j
            for i, j in self.jobs['interval'].items()
            if i != job_id
        }



from abc import ABCMeta, abstractmethod

from ddd.application.domain_registry import DomainRegistry


class EventAdapter(object, metaclass=ABCMeta):

    def __init__(self, log_service=None, service=None, listeners=None):
        super().__init__()

        self.log_service = \
            log_service \
            if log_service \
            else DomainRegistry.get_instance().log_service

        self.service = service
        self.listeners = [] if listeners is None else []
        self._assign_service_to_listeners()

    # Service

    def set_service(self, service):
        self.service = service
        self._assign_service_to_listeners()

    # Listeners

    def set_listeners(self, listeners):
        self.listeners = listeners
        self._assign_service_to_listeners()

    # Handling

    async def handle(self, message):
        for listener in self.listeners:
            if listener.listens_to(message['name']):
                await listener.handle(message)

    # Control

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass

    # Helpers

    def _assign_service_to_listeners(self):
        for listener in self.listeners:
            listener.service = self.service



import threading


class Transaction(object):
    def __init__(self):
        super(Transaction, self).__init__()

    async def __aenter__(self):
        return "dummy-transaction"

    async def __aexit__(self, type, value, traceback):
        pass


class Conn(object):
    def __init__(self):
        super(Conn, self).__init__()
        self.is_open = False

    def open(self):
        self.is_open = True

    def close(self):
        self.is_open = False

    def transaction(self):
        return Transaction()

    async def __aenter__(self):
        self.is_open = True

    async def __aexit__(self, type, value, traceback):
        self.is_open = False


class ConnectionHolder(object):
    def __init__(self):
        super(ConnectionHolder, self).__init__()

    async def __aenter__(self):
        self.conn = Conn()
        self.conn.open()
        return self.conn

    async def __aexit__(self, type, value, traceback):
        self.conn.close()


class MemoryPool(object):
    """
    The memory pool is used in the action decorator
    to create a "memory" transaction (for tests).
    """
    def __init__(self, max=80):
        super(MemoryPool, self).__init__()
        self.sem = threading.Semaphore(value=max)

    def acquire(self):
        self.sem.acquire()
        return ConnectionHolder()

    def release(self):
        self.sem.release()

    async def close(self):
        pass



from setuptools import find_packages
from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='ddd-for-python',
    version='0.9.3',
    author='David Runemalm, 2021',
    author_email='david.runemalm@gmail.com',
    description=
    'A domain-driven design (DDD) framework for Python.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/runemalm/ddd-for-python',
    project_urls={
        "Bug Tracker": "https://github.com/runemalm/ddd-for-python/issues",
    },
    package_dir={'': '.'},
    packages=find_packages(
        where='.',
        include=['ddd*',],
        exclude=[]
    ),
    license='GNU General Public License v3.0',
    install_requires=[
        'addict>=2.3.0',
        'aiokafka>=0.7.0',
        'APScheduler>=3.6.3',
        'arrow>=0.16.0',
        'asyncpg>=0.21.0',
        'azure-eventhub>=5.2.0',
        'azure-eventhub-checkpointstoreblob-aio>=1.1.1',
        'CMRESHandler>=1.0.0',
        'python-dotenv>=0.14.0',
        'slacker-log-handler>=1.7.1',
        'SQLAlchemy>=1.3.20',
        'uvloop>=0.14.0',
    ],
    tests_require=[
        'aiosonic>=0.7.2',
    ],
    python_requires='>=3.8.5',
)



from abc import ABCMeta, abstractmethod


class SchedulerAdapter(object, metaclass=ABCMeta):
    def __init__(self, exec_func, log_service):
        super().__init__()
        self.exec_func = exec_func
        self.log_service = log_service

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass

    @abstractmethod
    async def add_cron_job(
        self,
        id,
        name,
        args,
        kwargs,
        crontab_expr,
        weekdays,
        hour,
        minute,
        second,
        start_date,
        end_date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        pass

    @abstractmethod
    async def add_date_job(
        self,
        id,
        name,
        args,
        kwargs,
        date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        pass

    @abstractmethod
    async def add_interval_job(
        self,
        id,
        name,
        args,
        kwargs,
        interval,
        days,
        hours,
        minutes,
        seconds,
        start_date,
        end_date,
        timezone,
        coalesce=False,
        misfire_grace_time=(60 * 60),
        replace_existing=True,
    ):
        pass

    @abstractmethod
    async def get_all_jobs(self):
        pass

    @abstractmethod
    async def get_job(self, job_id):
        pass

    @abstractmethod
    async def get_job_count(self):
        pass

    @abstractmethod
    async def remove_job(self, job_id, raises=False):
        pass



import asyncio

from abc import ABCMeta, abstractmethod

from ddd.application.domain_registry import DomainRegistry


class EventPublisher(object, metaclass=ABCMeta):
    """
    The event publisher interface.
    """
    def __init__(self, log_service=None, keep_flushed_copies=False):

        self.log_service = \
            log_service \
            if log_service \
            else DomainRegistry.get_instance().log_service

        self.keep_flushed_copies = keep_flushed_copies

        self.published = {}
        self.flushed = {}

    # To override

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass

    @abstractmethod
    async def flush(self, event):
        """
        Flush (actually publish) the events.
        """
        pass

    # Methods

    async def publish(self, event):
        """
        Publish an event.
        """
        task_id = self._get_loop_task_id()

        if task_id not in self.published:
            self.published[task_id] = []

        # Log
        extra = {}

        self.log_service.debug(
            f"Publishing event: {event.name}",
            extra=extra,
        )

        self.published[task_id].append(event)

    def get_published(self):
        task_id = self._get_loop_task_id()

        if task_id in self.published:
            return self.published[task_id]

        return []

    def clear_published(self):
        task_id = self._get_loop_task_id()

        if task_id in self.published:
            self.published[task_id] = []

    def has_flushed(self, event, ignore_fields=None):
        ignore_fields = ignore_fields if ignore_fields is not None else []

        for task_id in self.flushed:
            for e in self.flushed[task_id]:
                if e.equals(event, ignore_fields=ignore_fields):
                    return True
        return False

    # Helpers

    def _get_loop_task_id(self):
        """
        Get currently executing asyncio event loop task.
        """
        if hasattr(asyncio, "current_task"):  # python >= 3.7
            task_id = id(asyncio.current_task())
        elif hasattr(asyncio.Task, "current_task"):  # python >= 3.6 < 3.7
            task_id = id(asyncio.Task.current_task())
        else:
            raise Exception(
                "Couldn't get currently running task on asyncio event loop. "
                "The version of python you're using isn't supported by "
                "this library."
            )



from addict import Dict

from unittest.mock import MagicMock

from ddd.domain.dummy.dummy import Dummy
from ddd.repositories.postgres.postgres_dummy_repository import \
    PostgresDummyRepository

from ddd.tests.dummy_action_test_case import DummyActionTestCase


class TestRepository(DummyActionTestCase):

    async def asyncSetUp(self):
        await super().asyncSetUp()

        self.repository = \
            PostgresDummyRepository(
                config=Dict({'database': {'dsn': "some-dsn"}}),
                db_service=None,
                log_service=self.deps.get_log_service(),
            )

        self.record_v1 = {
            'id': "some-dummy-id",
            'data': {
                'version': "1",
            }
        }

        self.record_v2 = {
            'id': "some-dummy-id",
            'data': {
                'version': "1",
                'second': "value",
            }
        }

        self.record_v3 = {
            'id': "some-dummy-id",
            'data': {
                'version': "1",
                'second': "value",
                'third': "value",
            }
        }

    async def test_get_versions_to_migrate_to_returns_unmigrated_versions_when_not_on_latest_version(
        self,
    ):
        # Setup
        Dummy.VERSION = "3"

        # Exercise
        versions = \
            self.repository._get_versions_to_migrate_to(
                record=self.record_v1,
                cls=Dummy,
            )

        # Assert
        self.assertEqual(
            range(2, 3),
            versions,
        )

    async def test_get_versions_to_migrate_to_returns_no_range_when_on_latest_version(
        self,
    ):
        # Setup
        Dummy.VERSION = "1"

        # Exercise
        versions = \
            self.repository._get_versions_to_migrate_to(
                record=self.record_v1,
                cls=Dummy,
            )

        # Assert
        self.assertEqual(
            range(0, 0),
            versions,
        )

    async def test_migrate_calls_all_migration_funcs(
        self,
    ):
        # Setup
        Dummy.VERSION = "3"

        # ..mock
        mock_migrate_v1_to_v2 = MagicMock(
            return_value=self.record_v2,
        )

        mock_migrate_v2_to_v3 = MagicMock(
            return_value=self.record_v3,
        )

        self.repository._migrate_v1_to_v2 = mock_migrate_v1_to_v2
        self.repository._migrate_v2_to_v3 = mock_migrate_v2_to_v3

        # Exercise
        self.repository._migrate(
            records=[self.record_v1],
            cls=Dummy,
        )

        # Assert
        mock_migrate_v1_to_v2.assert_called_with(
            record=self.record_v1,
        )

        mock_migrate_v2_to_v3.assert_called_with(
            record=self.record_v2,
        )



import os

from dotenv import find_dotenv

from ddd.tests.action_test_case import ActionTestCase

from ddd.application.dummy_application_service import DummyApplicationService
from ddd.utils.dep_mgr import DependencyManager
from ddd.utils.utils import read_config


class DummyActionTestCase(ActionTestCase):

    def __init__(self, methodName='runTest'):
        super().__init__(
            env_file_path=find_dotenv(filename=os.getenv('ENV_FILE')),
            methodName=methodName,
        )

    async def asyncSetUp(self):
        await super().asyncSetUp()

    async def asyncTearDown(self):
        await self.deps.get_dummy_repository().delete_all()

        await super().asyncTearDown()

    # Dependencies

    def read_config(self):
        self.config = read_config()

    def _create_deps_manager(self):
        self.deps = DependencyManager(
            config=self.config,
            loop=self.loop,
        )

    def get_service(self):
        if not self.service:
            self.service = \
                DummyApplicationService(
                    db_service=self.deps.get_db_service(),
                    domain_adapter=self.deps.get_domain_adapter(),
                    domain_publisher=self.deps.get_domain_publisher(),
                    event_repository=self.deps.get_event_repository(),
                    interchange_adapter=self.deps.get_interchange_adapter(),
                    interchange_publisher=
                    self.deps.get_interchange_publisher(),
                    job_service=self.deps.get_job_service(),
                    job_adapter=self.deps.get_job_adapter(),
                    log_service=self.deps.get_log_service(),
                    scheduler_adapter=self.deps.get_scheduler_adapter(),
                    dummy_repository=self.deps.get_dummy_repository(),
                    max_concurrent_actions=self.config.max_concurrent_actions,
                )

        return self.service

    # Actions

    # async def _create_user(
    #     self,
    #     user_id,
    #     created_at=arrow.utcnow(),
    #     updated_at=None,
    #     tags=None,
    #     username="some-username",
    #     email="some-email@nuvoair.com",
    #     first_name="John",
    #     last_name="Doe",
    #     token="some-token",
    #     make_request=False,
    # ):
    #     tags = tags if tags is not None else []
    #
    #     if make_request:
    #         raise NotImplementedError()
    #     else:
    #         command = \
    #             CreateUserCommand(
    #                 user_id=user_id,
    #                 created_at=created_at,
    #                 updated_at=updated_at,
    #                 tags=tags,
    #                 username=username,
    #                 email=email,
    #                 first_name=first_name,
    #                 last_name=last_name,
    #                 token=token,
    #             )
    #
    #         user = \
    #             await self.deps.service.create_user(
    #                 command=command,
    #             )
    #
    #         self.assertIsNotNone(user)
    #
    #         return user

    # Helpers

    def _disable_event_listeners(self):
        self._disable_domain_event_listeners()
        self._disable_integration_event_listeners()

    def _disable_domain_event_listeners(self, exclude=None):
        exclude = exclude if exclude is not None else []

        exclude = [
            e.__name__ for e in exclude
        ]

        for listener in self.deps.get_domain_adapter().listeners:
            if listener.listens_to_name not in exclude:
                listener.disable()
            else:
                listener.enable()

    def _disable_integration_event_listeners(self, exclude=None):
        exclude = exclude if exclude is not None else []

        exclude = [
            e.__name__ for e in exclude
        ]

        for listener in self.deps.get_interchange_adapter().listeners:
            if listener.listens_to_name not in exclude:
                listener.disable()
            else:
                listener.enable()



import asyncio
import os
import uvloop

from addict import Dict
from dotenv import find_dotenv, load_dotenv


class Config(object):
    """
    A class for holding the settings as defined by env file.

    Extend 'Config' if you want to add config settings for your context.

    We use 'addict' to be able to use dot-notation
    for referencing dict values.
    """
    TRUTHY = ["True", "true", "1", 1, True]

    def __init__(self, env_file_path=None):
        super().__init__()

        self._declare_settings()
        self._find_env_file(env_file_path)
        self._load_env_file()
        self._read_config()
        self._get_loop()

    def _declare_settings(self):
        """
        Declare all settings as 'None'.
        """
        self.env = None
        self.debug = None
        self.max_concurrent_actions = None
        self.top_level_package_name = None

        self.loop = Dict({
            'type': None,
            'instance': None,
        })

        self.http = Dict({
            'debug': None,
            'port': None,
        })
        self.database = Dict({
            'type': None,
            'postgres': Dict({
                'dsn': None,
            }),
        })
        self.auth = Dict({
            'full_access_token': None,
        })
        self.pubsub = Dict({
            'domain': Dict({
                'provider': None,
                'topic': None,
                'group': None,
            }),
            'interchange': Dict({
                'provider': None,
                'topic': None,
                'group': None,
            }),
            'kafka': Dict({
                'bootstrap_servers': None,
            }),
            'azure': Dict({
                'namespace': None,
                'namespace_conn_string': None,
                'checkpoint_store_conn_string': None,
                'blob_container_name': None,
            }),
        })
        self.jobs = Dict({
            'scheduler': Dict({
                'type': None,
                'dsn': None,
            })
        })
        self.slack = Dict({
            'token': None,
        })
        self.log = Dict({
            'enabled': None,
            'slack': Dict({
                'enabled': None,
                'errors': Dict({
                    'channel': None,
                })
            }),
            'kibana': Dict({
                'enabled': None,
                'url': None,
                'username': None,
                'password': None,
            }),
        })
        self.base_url = None
        self.tasks = Dict({
            'runner': Dict({
                'username': None,
                'password': None,
            }),
        })

    def _find_env_file(self, env_file_path):
        """
        Tries to find the env file path.

        Either as defined by 'env_file_path' or by searching recursively
        upwards from current workding dir.

        Throws error if no env file could be found, since it's required
        to launch the app.
        """
        if env_file_path in [None, ""]:
            env_file_path = \
                find_dotenv(
                    filename=os.getenv('ENV_FILE'),
                    raise_error_if_not_found=True,
                    usecwd=True,
                )
        self.env_file_path = env_file_path

    def _load_env_file(self):
        """
        Loads the environment file by loading the env vars defined in it
        so that they become accessible using 'os.getenv()'.
        """
        load_dotenv(
            dotenv_path=self.env_file_path,
        )

    def _read_config(self):
        """
        Read settings from env vars into this object's member variables.
        """
        self.env = os.getenv('ENV')
        self.debug = os.getenv("DEBUG")
        self.max_concurrent_actions = \
            int(os.getenv("MAX_CONCURRENT_ACTIONS", "10"))
        self.top_level_package_name = \
            os.getenv("TOP_LEVEL_PACKAGE_NAME")

        self.loop.type = os.getenv("LOOP_TYPE")

        self.http.debug = os.getenv("HTTP_DEBUG") in self.TRUTHY
        self.http.port = os.getenv("HTTP_PORT")

        self.database.type = os.getenv("DATABASE_TYPE")
        self.database.postgres.dsn = os.getenv("POSTGRES_DSN")

        self.auth.full_access_token = os.getenv("AUTH_FULL_ACCESS_TOKEN")

        self.pubsub.domain.provider = os.getenv("DOMAIN_PUBSUB_PROVIDER")
        self.pubsub.domain.topic = os.getenv("DOMAIN_PUBSUB_TOPIC")
        self.pubsub.domain.group = os.getenv("DOMAIN_PUBSUB_GROUP")

        self.pubsub.interchange.provider = \
            os.getenv("INTERCHANGE_PUBSUB_PROVIDER")
        self.pubsub.interchange.topic = \
            os.getenv("INTERCHANGE_PUBSUB_TOPIC")
        self.pubsub.interchange.group = \
            os.getenv("INTERCHANGE_PUBSUB_GROUP")

        self.pubsub.kafka.bootstrap_servers = \
            os.getenv("KAFKA_BOOTSTRAP_SERVERS")

        self.pubsub.azure.namespace = \
            os.getenv("AZURE_EVENT_HUBS_NAMESPACE")
        self.pubsub.azure.namespace_conn_string = \
            os.getenv("AZURE_EVENT_HUBS_NAMESPACE_CONN_STRING")
        self.pubsub.azure.checkpoint_store_conn_string = \
            os.getenv("AZURE_EVENT_HUBS_CHECKPOINT_STORE_CONN_STRING")
        self.pubsub.azure.blob_container_name = \
            os.getenv("AZURE_EVENT_HUBS_BLOB_CONTAINER_NAME")

        self.jobs.scheduler.type = os.getenv("JOBS_SCHEDULER_TYPE")
        self.jobs.scheduler.dsn = os.getenv("JOBS_SCHEDULER_DSN")

        self.slack.token = os.getenv("SLACK_TOKEN")

        self.log.enabled = os.getenv("LOG_ENABLED") in self.TRUTHY

        self.log.slack.enabled = \
            os.getenv("LOG_SLACK_ENABLED") in self.TRUTHY
        self.log.slack.errors.channel = \
            os.getenv("LOG_SLACK_CHANNEL_ERRORS")

        self.log.kibana.enabled = \
            os.getenv("LOG_KIBANA_ENABLED") in self.TRUTHY
        self.log.kibana.url = os.getenv("LOG_KIBANA_URL")
        self.log.kibana.username = os.getenv("LOG_KIBANA_USERNAME")
        self.log.kibana.password = os.getenv("LOG_KIBANA_PASSWORD")

        self.base_url = os.getenv('BASE_URL', 'empty-base-url')

        self.tasks.runner.username = os.getenv('TASK_RUNNER_USERNAME')
        self.tasks.runner.password = os.getenv('TASK_RUNNER_PASSWORD')

    def _get_loop(self):
        """
        Create the loop,
        add it to asyncio and this config.
        """
        loop = None

        if self.loop.type == "uvloop":
            loop = uvloop.new_event_loop()
        else:
            loop = asyncio.get_event_loop()

        asyncio.set_event_loop(loop)

        self.loop.instance = loop



from ddd.domain.event_publisher import EventPublisher


class MemoryEventPublisher(EventPublisher):
    """
    A memory event publisher.
    """
    def __init__(
        self,
        log_service=None,
        keep_flushed_copies=False,
    ):
        super().__init__(
            log_service=log_service,
            keep_flushed_copies=keep_flushed_copies
        )
        self.is_started = False
        self.primary_adapter = None

    # Flush

    async def flush(self, event):
        """
        Actually publishes the event,
        (called after request is done in the 'action' decorator).
        """
        if not self.is_started:
            raise Exception(
                "Can't flush the domain event because "
                "domain publisher isn't started."
            )

        task_id = self._get_loop_task_id()

        if self.primary_adapter:
            for listener in self.primary_adapter.listeners:
                if listener.listens_to(event.name):
                    try:
                        await listener.handle(event.serialize())
                    except Exception as e:
                        self.log_service.error(
                            "Listener failed to handle event '{}', "
                            "exception: {}".
                            format(
                                event.name,
                                str(e),
                            ),
                            exc_info=True,
                        )
        else:
            self.log_service.warning(
                "Couldn't flush events because no primary adapter is set."
            )

        if self.keep_flushed_copies:
            if task_id not in self.flushed:
                self.flushed[task_id] = []

            self.flushed[task_id].append(event)

    # Control

    async def start(self):
        self.is_started = True

    async def stop(self):
        self.is_started = False

    # Setters

    def set_primary_adapter(self, primary_adapter):
        self.primary_adapter = primary_adapter
